unit Main;

interface

uses
  System.SysUtils, System.Types, System.UITypes, System.Classes, System.Variants,
  FMX.Types, FMX.Controls, FMX.Forms, FMX.Graphics, FMX.Dialogs, FMX.StdCtrls,
  FMX.Layouts, FMX.Memo, FMX.ScrollBox, FMX.Controls.Presentation,
  Winsoft.Android.FTDI.D2XX, FMX.ListBox, FMX.Memo.Types;

type
  TFormMain = class(TForm)
    ButtonOpenClose: TButton;
    Memo: TMemo;
    ComboBoxDevices: TComboBox;
    Timer: TTimer;
    procedure ButtonOpenCloseClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure ComboBoxDevicesChange(Sender: TObject);
    procedure MemoChangeTracking(Sender: TObject);
    procedure TimerTimer(Sender: TObject);
  private
    { Private declarations }
    Device: TSerialDevice;
    Text: string;
    procedure AppendText(const NewText: string);
    procedure RefreshDevices;
    procedure Open;
    procedure Close;
  public
    { Public declarations }
  end;

var
  FormMain: TFormMain;

implementation

uses FMX.Helpers.Android, Androidapi.Jni, Androidapi.JNIBridge,
  Androidapi.Helpers, Androidapi.JNI.GraphicsContentViewText,
  Androidapi.JNI.JavaTypes;

{$R *.fmx}

procedure TFormMain.FormCreate(Sender: TObject);
begin
  Device := TSerialDevice.Create;
  RefreshDevices;
end;

procedure TFormMain.RefreshDevices;
begin
  ComboBoxDevices.Clear;
  Device.EnumDevices;
  for var I := 0 to Device.DeviceCount - 1 do
  begin
    const DeviceInfo = Device.DeviceInfo[I];
    try
      ComboBoxDevices.Items.Add(DeviceInfo.Description);
    finally
      DeviceInfo.Free;
    end;
  end;

  if ComboBoxDevices.ItemIndex = -1 then
  begin
    Close;
    if ComboBoxDevices.Items.Count > 0 then
      ComboBoxDevices.ItemIndex := 0;
  end;

  ComboBoxDevicesChange(Self);
end;

procedure TFormMain.ComboBoxDevicesChange(Sender: TObject);
begin
  ButtonOpenClose.Enabled := ComboBoxDevices.ItemIndex <> -1;
end;

procedure TFormMain.Open;
begin
  Memo.Lines.Clear;
  Text := Memo.Text;
  Memo.SetFocus;
  Memo.GoToTextEnd;
  Memo.SelStart := Length(Text);
  Memo.SelLength := 0;

  // show library version
  AppendText('FTDI library version: ' + IntToHex(Device.LibraryVersion, 8) + sLineBreak);

  // show device info
  const DeviceInfo = Device.DeviceInfo[ComboBoxDevices.ItemIndex];
  try
    AppendText('Device type: ' + DeviceInfo.DeviceType + sLineBreak);
    AppendText('Description: ' + DeviceInfo.Description + sLineBreak);
    AppendText('Serial number: ' + DeviceInfo.SerialNumber + sLineBreak);
  finally
    DeviceInfo.Free;
  end;

  // set communication parameters
  Device.DataBits := TDataBits._8;
  Device.StopBits := TStopBits._1;
  Device.Parity := TParity.None;
  Device.BaudRate := 9600;

  // open device
  Device.Open(ComboBoxDevices.ItemIndex);
  ComboBoxDevices.Enabled := False;
  ButtonOpenClose.Text := 'Close';
  Memo.Enabled := True;
  Timer.Enabled := True;
end;

procedure TFormMain.Close;
begin
  Timer.Enabled := False;
  Device.Close;
  ComboBoxDevices.Enabled := True;
  ButtonOpenClose.Text := 'Open';
  Memo.Enabled := False;
end;

procedure TFormMain.ButtonOpenCloseClick(Sender: TObject);
begin
  if not Device.Opened then
    Open
  else
    Close
end;

procedure TFormMain.MemoChangeTracking(Sender: TObject);
begin
  // Memo OnKeyDown doesn't work, so we need to track changes manually
  if Device.Opened then
  begin
    const MemoText = Memo.Text;
    if MemoText <> Text then
    begin
      if Length(MemoText) > Length(Text) then // character was added
      begin
        var KeyChar := #0;
        var Found := False;
        for var I := Low(Text) to High(Text) do
          if MemoText[I] <> Text[I] then
          begin
            KeyChar := MemoText[I]; // first different character found
            Found := True;
            Break;
          end;

        if not Found then
          KeyChar := MemoText[High(Text) + 1];

        Device.Write(Ord(KeyChar));
      end;
      Text := MemoText;
    end;
  end;
end;

procedure TFormMain.TimerTimer(Sender: TObject);
var Buffer: TArray<Byte>;
begin
  SetLength(Buffer, Device.InputCount);
  const Size = Device.Read(Buffer);
  if Size >= 1 then
  begin
    SetLength(Buffer, Size);
    AppendText(TEncoding.ANSI.GetString(Buffer));
  end;
end;

procedure TFormMain.AppendText(const NewText: string);
begin
  if NewText = '' then
    Exit;

  Text := Memo.Text + NewText;
  Memo.Text := Text;
  Memo.GoToTextEnd;
  Memo.SelStart := Length(Text);
  Memo.SelLength := 0;
end;

end.